from re import sub

def clean(text):
    text = text.lower()
    text = sub("[0-9]+", "number", text)
    text = sub("#", "", text)
    text = sub("\\n", "", text)
    text = text.replace('$', '@')
    text = sub("@[^\\s]+", "", text)
    text = sub("(http|https)://[^\\s]*", "", text)
    text = sub("[^\\s]+@[^\\s]+", "", text)
    text = sub('[^a-z A-Z]+', '', text)
    return text



import time
from threading import Thread
from collections import deque
from nltk.sentiment.vader import SentimentIntensityAnalyzer
import numpy as np

from tweepy import (OAuthHandler,
                    Stream)
from tweetfeels import (TweetData,
                        TweetListener)
from tweetfeels.utils import clean


class TweetFeels(object):
    """
    The controller.

    :param credentials: A list of your 4 credential components.
    :param tracking: A list of keywords to track.
    :param db: A sqlite database to store data. Will be created if it doesn't
               already exist. Will append if it exists.
    :ivar calc_every_n: Wont calculate new sentiment until there are n records
                        in the queue.
    :ivar lang: A list of languages to include in tweet gathering.
    """
    _db_factory = (lambda db: TweetData(db))
    _auth_factory = (
        lambda cred: OAuthHandler(cred[0], cred[1]
            ).set_access_token(cred[2], cred[3])
        )
    _listener_factory = (lambda ctrl: TweetListener(ctrl))
    _stream_factory = (lambda auth, listener: Stream(auth, listener))

    def __init__(self, credentials, tracking=[], db='feels.sqlite'):
        self._feels = TweetFeels._db_factory(db)
        _auth = TweetFeels._auth_factory(credentials)
        self._listener = TweetFeels._listener_factory(self)
        self._stream = TweetFeels._stream_factory(_auth, self._listener)
        self.tracking = tracking
        self.lang = ['en']
        self._sentiment = 0
        self._filter_level = 'low'
        self.calc_every_n = 10
        self._latest_calc = None
        self._tweet_buffer = deque()
        self.buffer_limit = 50

    def start(self, seconds=None):
        def delayed_stop():
            time.sleep(seconds)
            print('Timer completed. Disconnecting now...')
            self.stop()

        if len(self.tracking) == 0:
            print('Nothing to track!')
        else:
            self._stream.filter(
                track=self.tracking, languages=self.lang, async=True
                )
#  This does not work due to upstream bug in tweepy 3.5.0. They have fixed it in
#  https://github.com/tweepy/tweepy/pull/783
#            self._stream.filter(
#               track=self.tracking, languages=self.lang, async=True,
#               filter_level=self._filter_level
#               )
        if seconds is not None:
            t = Thread(target=delayed_stop)
            t.start()

    def stop(self):
        self._stream.disconnect()

    def on_data(self, data):
        """
        Note: Due to upstream bug in tweepy for python3, it cannot handle the
        `filter_level` parameter in the `Stream.filter` function. Therefore,
        we'll take care of it here. The problem has been identified and fixed
        by the tweepy team here: https://github.com/tweepy/tweepy/pull/783
        """
        filter_value = {'none': 0, 'low': 1, 'medium': 2}
        value = filter_value[data['filter_level']]

        if value >= filter_value[self._filter_level]:
            self._tweet_buffer.append(data)

            if len(self._tweet_buffer) > self.buffer_limit:
                t = Thread(target=self.clear_buffer)
                t.start()

    def clear_buffer(self):
        while True:
            try:
                # The insert calculates sentiment values
                self._feels.insert_tweet(self._tweet_buffer.popleft())
            except IndexError:
                break

    def on_error(self, status):
        self.start()

    @property
    def connected(self):
        return self._stream.running

    @property
    def sentiment(self):
        def avg_sentiment(df):
            avg = 0
            try:
                avg = np.average(
                    df.sentiment, weights=df.followers_count+df.friends_count
                    )
            except ZeroDivisionError:
                avg = 0
            return avg

        dfs = self._feels.tweets_since(self._latest_calc)
        for df in dfs:
            if(len(df)>self.calc_every_n):
                df = df.loc[df.sentiment != 0]  # drop rows having 0 sentiment
                df = df.groupby('created_at')
                df = df.apply(avg_sentiment)
                df = df.sort_index()
                for row in df.iteritems():
                    self._sentiment = self._sentiment*0.99 + row[1]*0.01
        return self._sentiment



import unittest
from unittest.mock import patch, MagicMock

from tweetfeels import TweetListener
from tweetfeels import Tweet

import json
from datetime import datetime


class Test_Listener(unittest.TestCase):
    def setUp(self):
        self.tweets_data_path = 'test/sample.json'
        self.disconnect_msg = """
            {
              "disconnect":{
                "code": 4,
                "stream_name":"",
                "reason":""
              }
            }
            """

    @patch('tweetfeels.TweetFeels')
    def test_listener(self, mock_feels):
        tl = TweetListener(mock_feels)
        with open(self.tweets_data_path) as tweets_file:
            lines = filter(None, (line.rstrip() for line in tweets_file))
            for line in lines:
                tl.on_data(line)
                mock_feels.on_data.assert_called()

    @patch('tweetfeels.TweetFeels')
    def test_on_disconnect(self, mock_feels):
        tl = TweetListener(mock_feels)
        tl.reconnect_wait = MagicMock()
        tl.on_disconnect(self.disconnect_msg)
        tl.reconnect_wait.assert_called_with('linear')
        tl._controller.start.assert_called_once()

    @patch('tweetfeels.TweetFeels')
    def test_on_connect(self, mock_feels):
        tl = TweetListener(mock_feels)
        tl.waited = 60
        tl.on_connect()
        self.assertEqual(tl.waited, 0)

    @patch('tweetfeels.TweetFeels')
    def test_on_error(self, mock_feels):
        tl = TweetListener(mock_feels)
        tl.reconnect_wait = MagicMock()
        tl.on_error(420)
        tl.reconnect_wait.assert_called_with('exponential')
        self.assertEqual(tl.waited, 60)
        mock_feels.on_error.assert_called_with(420)

    @patch('tweetfeels.TweetFeels')
    def test_reconnect_wait(self, mock_feels):
        tl = TweetListener(mock_feels)
        tl.waited = 0.1
        tl.reconnect_wait('linear')
        self.assertEqual(tl.waited, 1.1)
        tl.waited = 0.1
        tl.reconnect_wait('exponential')
        tl.reconnect_wait('exponential')
        self.assertEqual(tl.waited, 0.4)


class Test_Tweet(unittest.TestCase):
    def setUp(self):
        self.tweet_file = 'test/tweet.json'
        self.tweet = None
        with open(self.tweet_file) as twt_file:
            lines = list(filter(None, (line.rstrip() for line in twt_file)))
            self.tweet = Tweet(json.loads(lines[0]))

    def test_tweet_keys(self):
        self.assertEqual(self.tweet['followers_count'], 83)
        self.assertEqual(self.tweet['friends_count'], 303)
        self.assertTrue(len(self.tweet)>0)
        dt = datetime(2017, 2, 19, 19, 14, 18)
        self.assertEqual(self.tweet['created_at'], dt)

    def test_attributes(self):
        self.assertEqual(len(self.tweet), 33)
        self.assertTrue('followers_count' in self.tweet)
        self.assertTrue(isinstance(self.tweet['created_at'], datetime))

    def test_sentiment(self):
        self.assertEqual(self.tweet.sentiment['compound'], -0.2472)
        self.assertEqual(self.tweet.sentiment['pos'], 0.087)
        self.assertEqual(self.tweet.sentiment['neu'], 0.752)
        self.assertEqual(self.tweet.sentiment['neg'], 0.161)



#from distutils.core import setup
from setuptools import setup

filename = 'tweetfeels/version.py'
exec(compile(open(filename, "rb").read(), filename, 'exec'))

setup(name='tweetfeels',
      version=__version__,
      description='Real-time sentiment analysis for twitter.',
      author='Thomas Chen',
      author_email='tkchen@gmail.com',
      url='https://github.com/uclatommy/tweetfeels',
      download_url='https://github.com/uclatommy/tweetfeels/tarball/{}'.format(
          __version__
          ),
      packages=['tweetfeels'],
      classifiers=[
          'Development Status :: 4 - Beta',
          'License :: OSI Approved :: BSD License',
          'Natural Language :: English',
          'Programming Language :: Python :: 3.6',
          'Topic :: Scientific/Engineering :: Artificial Intelligence'
          ],
      install_requires=[
          'tweepy', 'h5py', 'nltk', 'numpy', 'oauthlib', 'pandas',
          'python-dateutil', 'pytz', 'requests', 'requests-oauthlib',
          'six', 'twython'
          ],
      test_suite='nose.collector',
      tests_require=['nose']
     )
